use std::fmt;
use std::fmt::Formatter;

fn capitalise(s: &str) -> String {
    let mut c = s.chars();
    match c.next() {
        None => String::new(),
        Some(f) => f.to_uppercase().collect::<String>() + c.as_str(),
    }
}

/// A basic register capable of storing binary data.
#[derive(Debug)]
pub struct BasicRegister {
    /// A [Vec] that stores the binary data of the register.
    memory: Vec<bool>,
}

impl BasicRegister {
    pub fn l_shift(&mut self) {
        let mut res: Vec<bool> = vec![false; self.memory.len()];

        res[..(self.memory.len() - 1)]
            .copy_from_slice(&self.memory[1..((self.memory.len() - 1) + 1)]);

        self.memory = res;
    }

    pub fn r_shift(&mut self) {
        let mut res: Vec<bool> = vec![false; self.memory.len()];

        for i in (1..self.memory.len()).rev() {
            res[i] = self.memory[i - 1];
        }

        self.memory = res;
    }

    pub fn reverse(&mut self) {
        self.memory = self.memory.iter().map(|val| !val).collect();
    }

    pub fn new(memory: Vec<bool>) -> Self {
        Self { memory }
    }
}

impl Default for BasicRegister {
    fn default() -> Self {
        Self {
            memory: vec![false],
        }
    }
}

impl fmt::Display for BasicRegister {
    fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
        write!(f, "Memory: [")?;

        for (count, v) in self.memory.iter().enumerate() {
            if count != 0 {
                write!(f, ", ")?;
            }
            write!(f, "{}", *v as u8)?;
        }

        write!(f, "]")
    }
}

impl fmt::Binary for BasicRegister {
    fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
        write!(f, "Memory: ")?;

        for v in self.memory.iter() {
            write!(f, "{}", *v as u8)?;
        }

        write!(f, "")
    }
}

fn read_vec(variable_name: &str) -> Vec<char> {
    loop {
        print!("Enter {variable_name}: ");
        let input: String = text_io::read!("{}\n");
        match input.chars().all(|c| c == '1' || c == '0') {
            true => return input.chars().collect(),
            false => eprintln!(
                "[ERROR] {} may contain only 1-s and 0-s with no whitespaces!",
                capitalise(variable_name)
            ),
        }
    }
}

pub fn char_to_bool_vector(char_vector: Vec<char>) -> Vec<bool> {
    let mut bool_vector: Vec<bool> = Vec::new();

    for value in char_vector.iter() {
        match value {
            '0' => bool_vector.push(false),
            '1' => bool_vector.push(true),
            _ => (),
        }
    }

    bool_vector
}

pub fn get_memory(variable_name: &str) -> Vec<bool> {
    char_to_bool_vector(read_vec(variable_name))
}
